/*
    Project: TCGINA
    Version: V1.23
    File   : random.c

    Copyright (c) 2005-2008 Author of TCGINA DLL. All rights reserved.
    Public Key Fingerprint: 294B A769 4A0A CC05 DAE6 00DD FF47 8C72 4097 67CE

    Covered by TrueCrypt License 2.3 the full text of which is contained
    in the file License.txt included in TCGINA binary and source code
    distribution archives.

    The copyright notice of this source code file and the public key
    fingerprint must be retained in full.
*/

#include "random.h"
#include <tomcrypt.h>
#include "secmem.h"
#include "misc.h"
#include <psapi.h>
#include <tchar.h>

#define RESEED_POOL_ROUNDS         30
#define RESEED_TIME                5000

#define RANDOM_POOL_SIZE           1031

typedef struct
{
    BOOL fFillRandomRequired;
    BYTE rgbRandomPool[RANDOM_POOL_SIZE];
    prng_state psState;
} PRNG, *PPRNG;

static PRNG *pPrng = NULL;

static PBYTE pbRandomPoolBegin;
static PBYTE pbRandomPoolEnd;
static PBYTE pbRandomPool;

static BOOL  fInitialSeedingRequired;
static DWORD dwNextSeedTime;
static DWORD dwRemainingPoolRounds;

BOOL (WINAPI * pfGetPerformanceInfo)(PPERFORMANCE_INFORMATION,
  DWORD) = NULL;

static BOOL CALLBACK RandomEnumWindowsProc(HWND hwnd, LPARAM lParam)
{
    TCHAR szTemp[512];
    WINDOWINFO wi;

    UNREFERENCED_PARAMETER(lParam);

    FillRandomPool(&hwnd, sizeof(hwnd));

    if (GetWindowText(hwnd, szTemp, countof(szTemp))) {
        FillRandomPool(szTemp, _tcslen(szTemp)*sizeof(TCHAR));
    }
    if (GetWindowModuleFileName(hwnd, szTemp, countof(szTemp))) {
        FillRandomPool(szTemp, _tcslen(szTemp)*sizeof(TCHAR));
    }
    if (GetWindowInfo(hwnd, &wi)) {
        FillRandomPool(&wi, sizeof(wi));
    }

    return TRUE;
}

static void AddSystemStatus(void)
{
    DWORD                   dwProcessCount;
    MEMORYSTATUS            ms;
    DWORD                   dw;
    LARGE_INTEGER           li;
    POINT                   pt;
    SYSTEMTIME              st;
    DWORD                   *pdwProcessIdBegin;

    EnumWindows(RandomEnumWindowsProc, 0);

    GlobalMemoryStatus(&ms);
    FillRandomPool(&ms, sizeof(ms));

    if (pfGetPerformanceInfo) {
        PERFORMANCE_INFORMATION pi;
        if (!pfGetPerformanceInfo(&pi, sizeof(pi))) {
            pi.ProcessCount = 512/sizeof(DWORD);
        }
        FillRandomPool(&pi, sizeof(pi));
        dwProcessCount = pi.ProcessCount * 2;
    } else {
        dwProcessCount = 1024;
    }
    pdwProcessIdBegin = malloc(dwProcessCount*sizeof(DWORD));
    if (pdwProcessIdBegin) {
        if (EnumProcesses(pdwProcessIdBegin, dwProcessCount*sizeof(DWORD),
            &dw)) {
            DWORD *pdwProcessId = pdwProcessIdBegin;
            DWORD *pdwProcessIdEnd = pdwProcessId + dw/sizeof(DWORD);

            FillRandomPool(pdwProcessId, dw);

            for (; pdwProcessId != pdwProcessIdEnd; pdwProcessId++) {
                HANDLE hProcess;
                HMODULE rghMod[1024];
                hProcess = OpenProcess(PROCESS_QUERY_INFORMATION |
                PROCESS_VM_READ,
                FALSE, *pdwProcessId);
                if (hProcess == NULL) {
                    hProcess = OpenProcess(PROCESS_QUERY_INFORMATION,
                        FALSE, *pdwProcessId);
                    if (hProcess == NULL) {
                        hProcess = OpenProcess(0, FALSE, *pdwProcessId);
                    }
                }
                if (hProcess) {
                    PROCESS_MEMORY_COUNTERS pmc;
                    IO_COUNTERS ic;
                    FILETIME ct;
                    FILETIME et;
                    FILETIME kt;
                    FILETIME ut;
                    if (GetProcessMemoryInfo(hProcess, &pmc,
                        sizeof(pmc))) {
                        FillRandomPool(&pmc, sizeof(pmc));
                    }
                    if (GetProcessIoCounters(hProcess, &ic)) {
                        FillRandomPool(&ic, sizeof(ic));
                    }
                    if (GetProcessTimes(hProcess, &ct, &et, &kt, &ut)) {
                        FillRandomPool(&ct, sizeof(ct));
                        FillRandomPool(&et, sizeof(ct));
                        FillRandomPool(&kt, sizeof(ct));
                        FillRandomPool(&ut, sizeof(ct));
                    }
                    if (EnumProcessModules(hProcess, rghMod,
                        sizeof(rghMod), &dw)) {
                        HMODULE *phTemp = rghMod;
                        HMODULE *phTempEnd = rghMod + dw/sizeof(HMODULE);
                        FillRandomPool(phTemp, dw);
                        for (; phTemp != phTempEnd; phTemp++) {
                            TCHAR szModuleName[MAX_PATH];
                            if (GetModuleFileNameEx(hProcess, *phTemp,
                            szModuleName, sizeof(szModuleName))) {
                                FillRandomPool(szModuleName,
                                _tcslen(szModuleName)*sizeof(TCHAR));
                            }
                        }
                    }
                    CloseHandle(hProcess);
                }
            }
        }
        free(pdwProcessIdBegin);
    }

    /* Time and counters */

    FillRandomPool(&dwNextSeedTime, sizeof(DWORD));

    dw = GetTickCount();
    FillRandomPool(&dw, sizeof(dw));

    QueryPerformanceCounter(&li);
    FillRandomPool(&li, sizeof(li));

    GetLocalTime(&st);
    FillRandomPool(&st, sizeof(st));

    dw = (DWORD)GetMessageTime();
    FillRandomPool(&dw, sizeof(dw));

    /* GUI */

    GetCursorPos(&pt);
    FillRandomPool(&pt, sizeof(pt));

    GetCaretPos(&pt);
    FillRandomPool(&pt, sizeof(pt));
}

#define LOAD_FUNCTION(h, x) \
  ((*(FARPROC*)&pf##x = GetProcAddress(h, #x)) != NULL)

BOOL InitRandom(void)
{
    if (pPrng == NULL) {
        HINSTANCE hDLL;
        LARGE_INTEGER li;
        DWORD         dw;
        HANDLE        h;
        SYSTEM_INFO   si;
        STARTUPINFO   sui;
        HCRYPTPROV    hProv;

        if ((hDLL = LoadLibrary(TEXT("PSAPI.DLL"))) != NULL) {
            LOAD_FUNCTION(hDLL, GetPerformanceInfo);
        }

        sec_heap_init();
        if ((pPrng = sec_malloc(sizeof(PRNG))) == NULL) {
            return FALSE;
        }
        pbRandomPool = pbRandomPoolBegin = pPrng->rgbRandomPool;
        pbRandomPoolEnd = pbRandomPoolBegin + RANDOM_POOL_SIZE;

        if (register_prng(&fortuna_desc) == -1) {
            DestroyRandom();
            return FALSE;
        }
        if (fortuna_start(&pPrng->psState) != CRYPT_OK) {
            DestroyRandom();
            return FALSE;
        }

        if (CryptAcquireContext(&hProv, NULL, MS_DEF_PROV, PROV_RSA_FULL,
          (CRYPT_VERIFYCONTEXT | CRYPT_MACHINE_KEYSET)) ||
          CryptAcquireContext(&hProv, NULL, MS_DEF_PROV, PROV_RSA_FULL,
          CRYPT_VERIFYCONTEXT | CRYPT_MACHINE_KEYSET | CRYPT_NEWKEYSET)) {
            CryptGenRandom(hProv, RANDOM_POOL_SIZE, pPrng->rgbRandomPool);
            CryptReleaseContext(hProv, 0);
        }

        /* System information */

        QueryPerformanceFrequency(&li);
        FillRandomPool(&li, sizeof(li));

        GetSystemInfo(&si);
        FillRandomPool(&si, sizeof(si));

        GetStartupInfo(&sui);
        FillRandomPool(&sui, sizeof(sui));

        /* Process information */

        dw = GetCurrentProcessId();
        FillRandomPool(&dw, sizeof(dw));

        h = GetCurrentProcess();
        FillRandomPool(&h, sizeof(h));

        dw = GetCurrentThreadId();
        FillRandomPool(&dw, sizeof(dw));

        h = GetCurrentThread();
        FillRandomPool(&h, sizeof(h));

        h = GetProcessHeap();
        FillRandomPool(&h, sizeof(h));

        fInitialSeedingRequired = TRUE;
    }

    return TRUE;
}

BOOL FillRandomPool(const void *pData, size_t cbDataSize)
{
    size_t         cbContiguousWriteSize;
    unsigned       *puDest;
    unsigned       *puDestEnd;
    const unsigned *puSrc;

    if (pPrng == NULL) {
        return FALSE;
    }

    if (cbDataSize > RANDOM_POOL_SIZE) {
        const BYTE *pbData = (const BYTE *)pData;
        while (cbDataSize) {
            size_t cbTemp = cbDataSize < RANDOM_POOL_SIZE ?
              cbDataSize : RANDOM_POOL_SIZE;
            if (!FillRandomPool(pbData, cbTemp)) {
                return FALSE;
            }
            pbData += cbTemp;
            cbDataSize -= cbTemp;
        }
        return TRUE;
    }

    cbContiguousWriteSize = (size_t)(pbRandomPoolEnd - pbRandomPool);
    if (cbContiguousWriteSize < cbDataSize) {
        if (!FillRandomPool(pData, cbContiguousWriteSize)) {
            return FALSE;
        }
        return FillRandomPool( ((const BYTE *)pData) + cbContiguousWriteSize,
          cbDataSize - cbContiguousWriteSize);
    }

    puSrc = (const unsigned *) pData;
    puDest = (unsigned *) pbRandomPool;
    pbRandomPool += cbDataSize & ~(sizeof(unsigned)-1);
    puDestEnd = (unsigned *) pbRandomPool;
    while (puDest != puDestEnd) {
        *puDest++ += *puSrc++;
    }

    cbDataSize &= sizeof(unsigned)-1;
    if (cbDataSize) {
        const BYTE *pbSrc = (const BYTE*) pData;
        BYTE       *pbDest = (BYTE *) pbRandomPool;
        BYTE       *pbDestEnd = pbDest + cbDataSize;
        while (pbDest != pbDestEnd) {
            *pbDest = (*pbDest + *pbSrc);
            pbSrc++;
            pbDest++;
        }
        pbRandomPool += cbDataSize;
    }

    if (pbRandomPool == pbRandomPoolEnd) {
        pbRandomPool = pbRandomPoolBegin;
        if (dwRemainingPoolRounds) {
            dwRemainingPoolRounds--;
        }
    }

    return TRUE;
}

BOOL Seed(void)
{
    PBYTE pbTemp;
    PBYTE pbTempEnd;

    AddSystemStatus();

    pbTemp    = pbRandomPoolBegin;
    pbTempEnd = pbTemp + (32 * 32);
    for (; pbTemp != pbTempEnd; pbTemp += 32) {
        if (fortuna_add_entropy(pbTemp, 32, &pPrng->psState) != CRYPT_OK) {
            return FALSE;
        }
    }
    if (fortuna_ready(&pPrng->psState) != CRYPT_OK) {
        return FALSE;
    }

    dwNextSeedTime = GetTickCount() + RESEED_TIME;
    dwRemainingPoolRounds = RESEED_POOL_ROUNDS;

    return TRUE;
}

static BOOL Reseed(void)
{
    PBYTE pbTempData = (PBYTE)sec_malloc(RANDOM_POOL_SIZE);

    if (pbTempData) {
        HCRYPTPROV    hProv;
        fortuna_read(pbTempData, RANDOM_POOL_SIZE, &pPrng->psState);
        FillRandomPool(pbTempData, RANDOM_POOL_SIZE);
        if (CryptAcquireContext(&hProv, NULL, MS_DEF_PROV, PROV_RSA_FULL,
            (CRYPT_VERIFYCONTEXT | CRYPT_MACHINE_KEYSET)) ||
            CryptAcquireContext(&hProv, NULL, MS_DEF_PROV, PROV_RSA_FULL,
            CRYPT_VERIFYCONTEXT | CRYPT_MACHINE_KEYSET | CRYPT_NEWKEYSET)) {
            memset(pbTempData, '\0', RANDOM_POOL_SIZE);
            CryptGenRandom(hProv, RANDOM_POOL_SIZE, pbTempData);
            FillRandomPool(pbTempData, RANDOM_POOL_SIZE);
            CryptReleaseContext(hProv, 0);
        }
        sec_free(pbTempData);
    }

    return Seed();
}

BOOL GetRandom(void *pDest, size_t cbDestSize)
{
    if (pPrng == NULL) {
        if (!InitRandom()) {
            return FALSE;
        }
    }
    if (fInitialSeedingRequired) {
        if (!Seed()) {
            return FALSE;
        }
        fInitialSeedingRequired = FALSE;
    } else if (GetTickCount() >= dwNextSeedTime && dwRemainingPoolRounds == 0) {
        if (!Reseed()) {
            return FALSE;
        }
    }

    return fortuna_read((unsigned char *)pDest, (unsigned long)cbDestSize,
      &pPrng->psState) == cbDestSize;
}

void DestroyRandom(void)
{
    if (pPrng) {
        fortuna_done(&pPrng->psState);
        sec_free(pPrng);
        pPrng = NULL;
    }
}
